need("mbase")
need("mgrid")
need("mgridio")

--- \brief Lua Wrapper for grid module
--- 
--- \file wrap_grid.lua
--- \author J.Fuhrmann, H.Langmach, T.Streckenbach
---

------------------------------------------------------------------------------
---
--- \brief Enhanced Grid constructor.
---
--- The (Lua only) call to Grid() can use
--- input from any of the stages of the 
--- \link wrap_gridgen.lua grid generation process \endlink  
--- and perform the remaining stages to create a grid.
--- 
--- Depending on the data in the description table, the following actions are taken:
--- - Create grid from tetgenio. The table needs to represent an instance of tetgenio.
--- - Create grid from output GridGenIO. If nproc is given, it will be partitioned.
--- - Create grid from input GridGenIO. If nproc is given, it will be partitioned.
--- - Create grid from Lua table which can be converted to  GridGenIO. If nproc is given, it will be partitioned.
--- - Create grid from array of arrays (coordinates,cells, cellregions, facets, facetregions).
---
--- \param table Description table.
--- \return Newly created grid.
---
--+ Grid *Grid(table);

function xprint(s)  end
function Grid(...)
   xprint "Grid"
   local desc=arg[1]
   local ext=arg[2]
   if not ext then ext={} end
   
   if not desc then 
   -- error("No description given!") 
      print("No description given!") 
      return nil
   end
-- Pass through
   if istype(desc,"Grid") then
      xprint "passtrough"
      return desc
      
-- create grid from tetgenio
   elseif desc[1] and istype(desc[1],"tetgenio") then
      xprint "tetgenio"
      return Grid(mgridio.ggCreateTGIO(desc[1]))
      
-- create grid from output GridGenIO. If nproc is given, it will be partitioned
   elseif desc.class=="gdesc" then
      xprint "gdesc"
      return Grid(GridGenIO(desc))
      
   elseif bool(desc.cells) then
      local cache=first(ext.cache,desc.cache,250)
      local nproc=first(ext.nproc,desc.nproc)
      if nproc then
	 xprint "gCreate"
         return mgrid.gCreate(desc.points,desc.cells,desc.cellregions,desc.facets,desc.facetregions,
                          nproc,cache)
      else
	 xprint "gCreateGG"
         return mgridio.gCreateGG(desc)
--(desc.points,desc.cells,desc.cellregions,desc.facets,desc.facetregions)
      end
      
-- create grid from input GridGenIO. If nproc is given, it will be partitioned
   elseif desc.flags and bool(desc.points) or istype(desc,"GridGenIO") then
      if istype(desc.points,"DArray2") then
         dim= mbase.da2NCol(desc.points)
      else
         dim=table.getn(desc.points[1])
      end
      local flags=first(ext.flags,desc.flags)
      local ggout
      if dim==1 then
         maxlen=desc.maxlen
         if not maxlen then maxlen=1 end
         ggout=gridgen1d(desc,maxlen,ext)
      elseif dim==2 then
         ggout=triangle(desc,flags,ext)
      elseif dim==3 then
         ggout=tetgen(desc,flags,ext)
      end

      local cache=first(ext.cache,desc.cache)
      local nproc=first(ext.nproc,desc.nproc)
      local ext={ flags=flags,
                  nproc=nproc,
                  cache=cache,
                 maxlen=maxlen}
	      xprint "recursive"
      return Grid(ggout,ext) -- call recursively
      
-- create grid from array of arrays
   elseif table.getn(desc)>0 then
      xprint "gCreate Array"
      return  mgrid.gCreate(
                      DArray2(desc[1]),
                      IArray2(desc[2]),
                      IArray1(desc[3]),
                      IArray2(desc[4]),
                      IArray1(desc[5]))
   end
   
end


--Grid.addref=gAddref
--getmetatable(Grid)["__call"]=callnew



------------------------------------------------
---
--- \brief Coord search  (very slow)
---
--- \param g Grid input
--- \param p Search for nodenumber of coord p
--- \param eps Coord-element equal
--- \return Zero if not found or nodenumber of coord p
--+ int gFindNode(Ptr<Grid> g, Ptr<DArray1> p,double eps);

function gFindNode(g,p,eps)
 local coords=gCoord(g)
 local i,j,d,found,c
 
 for i=1,gNElements(g,gNode) do
   found=true
   c=coords[i]
   for j=1,table.getn(p) do
     d=math.abs(c[j]-p[j])
     if d>eps then found=false ; break end
   end
   if found then return i end
  end
  
  return 0
end

------------------------------------------------
---
--- \brief Refine on points
---
--- \param grid Grid input
--- \param refpoints additional points
--- \param flags Refine flags for grid generation
--- \param eps Coord-element equal (optional)
--- \return Refined grid. If eps is not nil, nodenumbers of refpoints in the
---         refined grid are also returned.
--+ Ptr<Grid> gRefineF(Ptr<Grid> g, Ptr<DArray2> refpoints, char* flags, double eps);

function gRefineF(grid,refpoints,flags,eps)
  local ggg=Grid(grid)
  local gg=ggCreateG(ggg)
  local ngeo
  
  -- prepare GridGenIO
  if type(refpoints) == "table" then
   
   ngeo=table.getn(refpoints)
   local dim=gGridDim(grid)
   gg.refpoints=da2Create(ngeo,dim)
   
   local p,d
   for p=1,ngeo do
    for d=1,dim do
     gg.refpoints[p][d]=refpoints[p][d]
    end
   end
   
  else
   ngeo=da2NRow(refpoints)
   gg.refpoints=refpoints
  end
  
  -- refine
  local rgrid=gGenerate(gg,flags)
  if not eps then
   return rgrid
  end
  
  -- search node numbers
  local nnrs={}
  for p=1,ngeo do
   nnrs[p]=gFindNode(rgrid,refpoints[p],eps)
  end
  
  return rgrid,nnrs
end


------------------------------------------------------------------------------
---
--- \brief Create an instance of gdesc (Grid description table)
---
--- \param eps Identify points  closer than \p eps. Optional.
--+ gdesc gdescCreate( double eps);



---
--- \brief Lua class gdesc.
---
--- Create description table step by step.
--- This should ease the process of grid description considerably.
---
--+ class gdesc { public:

function gdescCreate(_eps)
   local g={}
   g.class="gdesc"
   g.eps=_eps or 1e-7  -- epsilon to identify points
   g.points={}  -- points (not necessarily identified)
   g.facets={}  -- facet list
   g.facetregions={} 
   g.regionpoints={}
   g.regionvolumes={}
   g.regionnumbers={}
   g.nproc=1
   g.cache=256
   g.flags="pqAaQ"
                
   g._facetregion=1    -- actual facet region
   g._regionvolume=0.1 -- actual region volume
   g._regionnumber=1   -- actual region numbert


--- \brief Insert point
---
--- Insert point (tuple of coordinates) into
--- point list and return its index in order to be used
--- in facet definitions.
--- \param pt Lua table containing point coordinates
--- \return point number in description table
--+ int point(pt);

   function g:point(pt)
      local i,n
      local ptx={}
      n = table.getn(pt)
      for i=1,n do 
        ptx[i]=pt[i] 
      end
      n = 1 + table.getn(self.points) -- new point number
      --table.insert(self.points,ptx)
      self.points[n] = ptx
      if self.debug then
         print("points[",n,"]=",table.concat(ptx,", "))
      end
      return n
   end
   
--- \brief Set actual facet region
---
--- Set facet region number for subsequent facets
--- 
--- \param reg Facet region
--+ facetregion(reg);
   
   function g:facetregion(reg)
      self._facetregion=reg
   end
   
--- \brief Insert facet.
---   
--- Define facet either by a tuple of point numbers
--- or by a tuple of points which are inserted point lists
--- and whose thus obtained indices then create a point number
--- tuple.
--- \param pts Lua table containing either the numbers of the points
--- the face is made off, or their coordinates. These have to be coplanar.

--+ facet(fac);
   function g:facet(fac)
      local myfac,n
      if type(fac[1])=="number" then
         myfac=fac
      elseif type(fac[1])=="table" then
         myfac={}
         n = table.getn(fac)
         for i=1,n do
            myfac[i]=self:point(fac[i])
         end
      else return 0      
      end
      -- return if there are identical points in the list myfac 
      local i,j
      n = table.getn(myfac)  
      for i=1,n do 
        for j=i+1,n do 
          if myfac[i]==myfac[j] then return 0 
          end 
        end
      end       
      -- insert the facet
      n = 1 + table.getn(self.facets) -- new facet number
      self.facets[n] = myfac
      self.facetregions[n] = self._facetregion
      return n   
   end
   
--- \brief Set region number
--- 
--- Set region number for the next region points
--- \param num Region number
--+ regionnumber(num);

   function g:regionnumber(num)
      self._regionnumber=num
   end

--- \brief Set actual region element volume.
---
--- Set the maximal volume for elements for the next region points
--- \param vol Max. element volume
--+ regionvolume(vol);

   function g:regionvolume(vol)
      self._regionvolume=vol
   end

--- \brief Set region point for actual region number.
---
--- \param pt Lua   table containing coordinates
--+ regionpoint(pt);

   function g:regionpoint(pt)
      table.insert(self.regionpoints,pt)
      table.insert(self.regionnumbers,self._regionnumber)
      table.insert(self.regionvolumes,self._regionvolume)
   end
   
--- \brief Partitioning: maximum number of elements in grid zone
---
--- Default value: 256
--+ int ncache;

--- \brief Partitioning: Number of processors.
---
--- Default value: nil (Dummy partitioning for 1 processor)
--+ int nproc;
---

--- \brief Grid generator control flags
---
--+ char *flags;


--- \brief Prepare generation.
--- 
---+ prepare_gridgenio();
   function g:prepare_gridgenio()
      -- Try to find a region point if none had been
      -- defined: use the average.
      local dim
      if table.getn(self.regionpoints)==0 then
        
         local i, idim,pt,np
         dim=table.getn(self.points[1])
         pt={}
         np=table.getn(self.points)
         
         for idim=1,dim do pt[idim]=0.0 end
         for i=1,np do
            for idim=1,dim do pt[idim]=pt[idim]+self.points[i][idim] end
         end
         for idim=1,dim do pt[idim]=pt[idim]/np end
         self:regionpoint(pt)
      end
   end


--- \brief Generate the grid.
--- 
---  Generate the grid  corresponding to the actual state
--- of the description.
---+ generate();
   function g:generate() -- for backward compatibility
      return Grid(self)
   end

   return g
end

--+ };

------------------------------------------------------------------------------
--- 
--- \brief Enhanced features for two-dimensional grid description
---        
---
--- \brief Create an instance of pgdesc (planar grid description)
---
--- \param eps Identify points  closer than \p eps
--+ gdesc pgdescCreate( double eps);



---
--- \brief Lua class pgdesc.
---
--- Create a planar grid description step by step
--- using incremental generation of curves. 
--- This works as easy as drawing lines with a pencil.
---
--+ class pgdesc { public:


function pgdescCreate(_eps)


   local g= gdescCreate(_eps)
   
   -- g.class="pgdesc"    
   g.class="gdesc"     -- pgdesc can be seen as a derived class of gdesc 

   g._savedpoint=0     -- actual point for moveto, lineto, ...  
   -- TODO: for splines we need a stack of points

   g._edgelength=0.1  -- edge length for curve approximation; 0 means "undefined" 

  --- \brief Defines an edge length for the approximation of subsequent curves. 
  --- \param h Edge length. 
  --+ edgelength(h); 
  function g:edgelength(h)
    self._edgelength = h
  end
  
  --- \brief  Find a point in the table of points.
  ---
  --- \param  pt Table of coordinates.
  --- \param  nlast Number of points to be included into the search. Optional.
  ---         If this parameter is not nil, then the point pt
  ---         will be compared with the last nlast points in the point list,
  ---         otherwise with all points.
  --- \return Index of the point or 0.
  ---
  --+ int findpoint(pt, nlast);
         
  function g:findpoint(pt,nlast)
    local n= table.getn(self.points)
    if n<1 then return 0 
    end
    local start = 1
    if nlast then start = math.max(1,n-nlast+1) 
    end
    local dim = math.min(table.getn(self.points[n]),table.getn(pt))
    local i,p,pt1,found
    for p=start,n do
      pt1 = self.points[p]
      found = true
      for i=1,dim do 
        if math.abs(pt1[i]-pt[i]) > self.eps then 
          found=false 
          break
        end 
      end--for i
      if found then
        return p
      end 
    end--for p
    return 0
  end--function findpoint

  --- \brief  Append a new point to the table of points.
  ---         (Remark: replaces  g.point(). TODO: unify both methods.)
  --- \param  pt Table of coordinates.
  --- \return Index of the new point.
  --+ int newpoint(pt);

   
  function g:newpoint(pt)
      local p = 0
      
      -- compare with existing points
      p = self:findpoint(pt)  -- compare pt with existing points
      if p>0 then 
        if self.debug then 
          print("newpoint: found [",p,"]=={", table.concat(pt,", "),"}")
        end
        return p
      end
      
      -- insert the point if it was not found 
      p = 1 + table.getn(self.points)   -- new point index 
      
      if self.debug then 
        print("newpoint: [",p,"]={", table.concat(pt,", "),"}")
      end

      self.points[p] = pt      -- insert the new point
      return p                 -- return the point index
  end--function newpoint


  --- \brief  Append a new facet to the table of facets.
  ---         (Remark: replaces  g.facet(). TODO: unify both methods.)
  --- \param  fp Table of facet point indices.
  --- \return Index of the new facet or 0.
  ---
  --+ int newfacet(fp);

  function g:newfacet(fp)
      -- return if there are identical points in the list myfac 
      local i,j
      local n = table.getn(fp)  
      for i=1,n do 
        for j=i+1,n do 
          if fp[i]==fp[j] then 
            print("newfacet: [","?","]={", table.concat(fp,", "),"}") 
            error("newfacet: ERROR, identical points exist") 
            return 0 
          end 
        end
      end
      -- insert new facet       
      local f = 1 + table.getn(self.facets) -- new facet index
      if self.debug then 
        print("newfacet: [",f,"]={", table.concat(fp,", "),"}") 
      end
      self.facets[f] = fp
      self.facetregions[f] = self._facetregion
      return f
  end--function newfacet


--[[
  function g:pushpoint(pt)
    if not self._pointstacktop then self._pointstacktop = 0 
    end
    self._pointstacktop = self._pointstacktop+1
    self._pointstack[self._pointstacktop] = pt
  end

  function g:poppoint(pt)
    if not self._pointstacktop then error("poppoint: no stack found") 
    end
    if self._pointstacktop <= 0 then return 0
    end
    self._pointstacktop = self._pointstacktop-1
    return self._pointstack[self._pointstacktop+1]
  end

  function g:getsavedpoint()
    local p1 = table.getn(self.points)  -- last index in table points
    if self._savedpoint and self._savedpoint>0 and self._savedpoint<p1 then 
      p1 = self._savedpoint 
    end
    pt1 = self.points[p1]
    return p1,pt1
  end
]]

  --- \brief  Move the (virtual) pen to the target point pt.
  --- \param  pt Target point.  
  ---         pt is either an existing point index or a table of point coordinates.
  ---         In the latter case, the point is added to the table "g.points".
  --- \return Index of the target point.
  --+ int moveto(pt);
  function g:moveto(pt)
    local p2,pt2   -- index and coordinates of the target point
    if type(pt)=="table" then -- the argument is a table of coordinates
      pt2 = pt;  
      p2  = self:newpoint(pt2)
    elseif type(pt)=="number" then -- the argument is an existing point index
      p2  = pt 
    else
      error("moveto: no valid target point")
      return 0  
    end
    if self.debug then 
      pt2 = self.points[p2]
      print("moveto:   [",p2,"]={", table.concat(pt2,", "),"}") 
    end
    self._savedpoint = p2  -- save the point index to mark the pen position
    return p2
  end--function moveto



  --- \brief  Generate a line from the current pen position to the target point pt.
  ---         This function works only in 2d ! 
  --- \param  pt Target point. 
  ---         pt is either an existing point index or a table of point coordinates.
  ---         In the latter case, the point is added to the table "g.points".
  --- \return Index of the target point.
  --+ int lineto(pt); 
  --- Example:
  ---   
  ---   g:moveto{0,0}
  ---   g:facetregion(1);  p = g:lineto{1,0}  
  ---   g:facetregion(2);  g:lineto{1,1}  
  ---   g:facetregion(3);  g:lineto{0,1}  
  ---   g:facetregion(4);  g:lineto(p)  

  function g:lineto(pt)
    local p1 = table.getn(self.points)  -- last index in table points
    if self._savedpoint and self._savedpoint>0 and self._savedpoint<p1 then 
      p1 = self._savedpoint 
    end
    
    local p2,pt2    -- index and coordinates of the target point
    if type(pt)=="table" then -- the argument is a table of coordinates
      pt2 = pt;  
      p2  = self:newpoint(pt2)
    elseif type(pt)=="number" then -- the argument is an existing point index
      p2  = pt 
    else
      error("lineto: no valid target point") 
      return 0 
    end
    if self.debug then 
      pt2 = self.points[p2]
      print("lineto:   [",p2,"]={", table.concat(pt2,", "),"}") 
    end
    
    if p1>0 then self:newfacet{p1,p2}
    end
    self._savedpoint = p2   -- save the point index to mark the pen position
    return p2
  end--function lineto


  --- \brief  Generate a Bezier curve from the current pen position 
  ---         to the target point pt.
  ---         This function works only in 2d ! 
  --- \param  pt Target point. 
  ---         pt is either an existing point index or a table of point coordinates.
  ---         In the latter case, the point is added to the table "g.points".
  --- \param  ptaux Table of aux. points. 
  --- \param  nedges Number edges approximating the curve.
  ---         If this parameter is omitted, then it will be determined from
  ---         a proposed edgelength, cf. method edgelength(). 
  --- \return Index of the target point.
  --+ int bcurveto(pt,ptaux,nedges);
  
  function g:bcurveto(pt,ptaux,nedges)

    local p1 = table.getn(self.points)  -- last index in table points
    if self._savedpoint and self._savedpoint>0 and self._savedpoint<p1 then 
      p1 = self._savedpoint 
    end
    if p1>0 then pt1 = self.points[p1]
    else
      return self:moveto(pt) 
    end

    local p2,pt2    -- index and coordinates of the target point
    if type(pt)=="table" then -- the argument is a table of coordinates
      pt2 = pt;  
      p2  = self:newpoint(pt2)
    elseif type(pt)=="number" then -- the argument is an existing point index
      p2  = pt 
      pt2 = self.points[p2]
    else
      error("bcurveto: no valid target point")  
      return 0
    end--if
    if self.debug then 
      print("bcurveto: [",p2,"]={", table.concat(pt2,", "),"}") 
    end--if

    if not ptaux then ptaux={}
    elseif type(ptaux)=="number" then ptaux={}
    elseif type(ptaux[1])=="number" then ptaux={ptaux} 
    end
    local i,j,p,pold
    local n = 1+table.getn(ptaux)
    local t = 0
    local c,m
    local ptx={}
    if nedges then m = math.abs(nedges)
    else
      m = math.sqrt((pt2[1]-pt1[1])^2+(pt2[2]-pt1[2])^2)/(self._edgelength)
    end     
    local h = 1/math.ceil(m)

    p = p1
    for i=1,m-1 do
      t = t+h
      c = math.pow(1-t,n)
      ptx[1] = pt1[1]*c
      ptx[2] = pt1[2]*c
      for j=1,n-1 do
        c = c*(n-j+1)/j*t/(1-t)
        ptx[1]= ptx[1]+ptaux[j][1]*c
        ptx[2]= ptx[2]+ptaux[j][2]*c
      end
      c=c/n*t/(1-t)
      ptx[1]= ptx[1]+pt2[1]*c
      ptx[2]= ptx[2]+pt2[2]*c

      pold = p
      p = self:newpoint{ptx[1],ptx[2]}
      self:newfacet{pold,p}     
    end
    pold = p
    self:newfacet{pold,p2}
    self._savedpoint = p2   -- save the point index to mark the pen position
    return p2
  end--function bcurveto



  --- \brief  Generate an arc (of a circle sector) which goes from the current 
  ---         pen position to the target point pt.
  ---         This function works only in 2D.
  --- \param  pt Target point. 
  ---         pt is either an existing point index or a table of point coordinates.
  ---         In the latter case, the point is added to the table "g.points".
  --- \param  angle Angle of the arc in degrees. 
  ---         If the angle is positive, then the arc is drawn counter-clockwise
  ---         (i.e. in mathematical positive sense). 
  ---         If the angle is negative, then the arc is drawn clockwise.
  ---         The range of the angle is 0 < abs(angle) < 360.
  ---         If the angle is near to 0, then the radius of the circle tends to infinite.
  ---         In this case, a straight line is generated. 
  ---         Similar, if abs(angle) is 360 degrees, then the radius of the circle 
  ---         becomes also infinite. But in this case we jump to the target point
  ---         without generating a circle. 
  --- \param  nedges Number edges approximating the curve.
  ---         If this parameter is omitted, then it will be determined from
  ---         a proposed edgelength, cf. method edgelength(). 
  --- \return Index of the target point.
  --+ int arcto(pt,angle,nedges);
  --
  --  TODO:   unification of different possibilities to draw an arc
  --          g:arcto(pt,{a=angle, r=radius, via={x,y},  tangent={dx,dy}, n=number, ...} )
  --
  function g:arcto(pt,angle,nedges)
   --
   --              center * 
   --                    /|\  b = angle/2                    
   --                   / |b\                    
   --        radius  r /  |  \                  
   --                 /   |L  \r   r = radius = (D/2)/sin(angle/2)
   --                /    |    \                 
   --  source point o-----+-----o  target point pt2  
   --           pt1  \     D/2 /   
   --                 '-.___.-'  D = distance between source and target point
   --    
   --                            L = r*cos(angle/2) = (D/2)*cot(angle/2)   
   --
   --  Let be i the inclination angle of the vector  pt2 - pt1,
   --  i.e.  pt2 - pt1 = { D*cos(i), D*sin(i) } 
   --  then 
   --  center = (pt2+pt1)/2 + { - L*sin(i), L*cos(i) }
   --
   -- determine the source point
   
   local p1 = table.getn(self.points)  -- last point index 
   if self._savedpoint and self._savedpoint>0 and self._savedpoint<p1 then 
     p1 = self._savedpoint 
   end
   if p1<1 then -- there is no source point
      return self:moveto(pt)
   end
   local pt1 = self.points[p1]  -- coordinates of the source point


   local PI = 3.14159265358979323846264
   local a

   if bool(angle) then     a = PI*angle/180  -- angle in rad 
                  else     a = 0
   end

   -- special cases

   if math.abs(a) < self.eps then  -- insert a straight line
      return self:lineto(pt)
   end
   if math.abs(a) + self.eps > 2*PI then  -- jump
      return self:moveto(pt)
   end
   
    local p2,pt2    -- index and coordinates of the target point
    if type(pt)=="table" then -- the argument is a table of coordinates
      pt2 = pt;  
      p2  = self:newpoint(pt2)
    elseif type(pt)=="number" then -- the argument is an existing point index
      p2  = pt 
      pt2 = self.points[p2]
    else
      error("arcto:    no valid target point")  
      return 0
    end--if
    if self.debug then 
      print("arcto:    [",p2,"]={", table.concat(pt2,", "),"}") 
    end--if
   
   local distance = math.sqrt( (pt1[1]-pt2[1])^2 + (pt1[2]-pt2[2])^2 )
      
   if distance <= self.eps then -- insert a line
     self:newfacet{p1,p2}
     self._savedpoint  = p2
     return p2
   end
     
   local cosi = (pt2[1]-pt1[1])/distance  -- cos of inclination angle
   local sini = (pt2[2]-pt1[2])/distance  -- sin of inclination angle
   
   local r = 0.5*distance/math.sin(a/2)   -- !!! may be negative
   local L = r*math.cos(a/2)              -- !!! may be negative
   
   local center = {(pt1[1]+pt2[1])/2 - L*sini, (pt1[2]+pt2[2])/2 + L*cosi} 
   
   local m
   if bool(nedges) then
       m = math.ceil(nedges)
   elseif self._edgelength>0 then
       m = math.ceil(math.abs(a*r/self._edgelength))  
   else
       m = math.ceil(math.abs(a*18/PI))       
   end     
   m = math.max(1,m)

   local dt = a/m                          -- !!! may be negative             
   local p = p1
   local t = -PI/2-a/2
   local p_old,i,v,pti

   for i=1,m-1 do
      t = t+dt
      v = { r*math.cos(t), r*math.sin(t) } -- !!! here, r may be negative
      pti = {cosi*v[1] - sini*v[2], 
             sini*v[1] + cosi*v[2]}  -- rotation  
      p_old = p
      p = self:newpoint({center[1] + pti[1], center[2] + pti[2]})
      self:newfacet{p_old,p}
   end

   self:newfacet{p,p2}
   self._savedpoint  = p2
   return p2
  end -- function arcto 



  --- \brief  Generate an arc (of a circle sector) from the current 
  ---         pen position to the target point pt using an auxiliary point.
  ---         This function works only in 2D.
  --- \param  pt target point. 
  ---         pt is either an existing point index or a table of point coordinates.
  ---         In the latter case, the point is added to the table "g.points".
  --- \param  ptaux auxiliary point. 
  ---         ptaux is either an existing point index or a table of point coordinates.
  ---         In the latter case, the point is added to the table "g.points".
  --- \param  nedges Proposed number edges approximating the curve. The minimum is 2.
  ---         If this parameter is omitted, then its value will be determined from
  ---         a proposed edgelength, cf. method edgelength(). 
  --- \return index of the target point.
  ---
  --+ int arc3to(pt,ptaux,nedges,debug);
  --
  function g:arc3to(pt,ptaux,nedges,debug)
   --
   --                o target point
   --               / '.  arc part 2
   --              /a2  .
   --     center  *-----o auxilary point
   --              \a1  '
   --        radius \ .'  arc part 1   
   --                o source point
   --
   local p1 = table.getn(self.points)  -- index of source point 
   if self._savedpoint and self._savedpoint>0 and self._savedpoint<p1 then 
     p1 = self._savedpoint 
   end
   if p1<1 then -- there is no source point
     self:moveto(ptaux)
     return self:lineto(pt)
   end
   local pt1 = self.points[p1]  -- coordinates of the source point
   
   local p3,pt3    -- index and coord. of the auxilary point between source and target
      
   if type(ptaux)=="table" then -- the argument is a table of coordinates
      pt3 = ptaux  
      p3  = self:newpoint(pt3)
   elseif type(ptaux)=="number" then -- the argument is an existing point index
      p3  = ptaux 
      pt3 = self.points[p3]
   else
      error("arc3to:   no valid aux. point")  
      return 0
   end--if
   if self.debug then 
      print("arc3to: aux. point [",p3,"]={", table.concat(pt3,", "),"}") 
   end--if
         
   local p2,pt2    -- index and coordinates of the target point
   if type(pt)=="table" then -- the argument is a table of coordinates
      pt2 = pt;  
      p2  = self:newpoint(pt2)
   elseif type(pt)=="number" then -- the argument is an existing point index
      p2  = pt 
      pt2 = self.points[p2]
   else
      error("arc3to:   no valid target point")  
      return 0
   end--if
   if self.debug then 
      print("arc3to:   [",p2,"]={", table.concat(pt2,", "),"}") 
   end--if
   
   local v1 = {pt1[1]-pt3[1],pt1[2]-pt3[2]}   -- vector v1 = (pt1 - pt3)
   local v2 = {pt2[1]-pt3[1],pt2[2]-pt3[2]}   -- vector v2 = (pt2 - pt3)

   local d1sq = ( v1[1]^2 + v1[2]^2 )         -- square( length (v1) )
   local d2sq = ( v2[1]^2 + v2[2]^2 )         -- square( length (v2) )


   if d1sq <= self.eps or d2sq <= self.eps then 
     self:newfacet{p1,p3}
     self:newfacet{p3,p2}
     self._savedpoint  = p2
     return p2
   end        
   
   -- We determine the vector c = (center - pt3) 
   -- as solution of the following linear system:
   --
   --   v1[1]*c[1] + v1[2]*c[2] = d1sq/2
   --   v2[1]*c[1] + v2[2]*c[2] = d2sq/2
   
   local det  = v1[1]*v2[2] - v1[2]*v2[1]     
      
   if math.abs (det) < self.eps^2 then -- insert a line
     self:newfacet{p1,p3}
     self:newfacet{p3,p2}
     self._savedpoint  = p2
     return p2
   end
   
   local c = { ( v2[2]*d1sq - v1[2]*d2sq)*0.5/det,
               (-v2[1]*d1sq + v1[1]*d2sq)*0.5/det }

   local center = {c[1]+pt3[1], c[2]+pt3[2]}
   local rsq    = (c[1]^2+c[2]^2)              
   local r      = math.sqrt(rsq)              

   local PI = 3.14159265358979323846264
   
   local function math_acos(X) 
     -- if X<-1 then return math.pi end
     if X<-1 then return PI end
     if X> 1 then return 0 end
     return math.acos(X)
   end      
   
   local a1 = math_acos(1-(c[1]*v1[1]+c[2]*v1[2])/rsq)  -- 1st angle in rad
   local a2 = math_acos(1-(c[1]*v2[1]+c[2]*v2[2])/rsq)  -- 2nd angle in rad

   local s1 = -(v1[1]*c[2]-v1[2]*c[1])  -- (v1-c) X (-c)  == {0,0,r^2 * sin(a1)}
   local s2 =  (v2[1]*c[2]-v2[2]*c[1])  -- (-c) X (v2-c)  == {0,0,r^2 * sin(a2)}

   if s1 < 0 then a1 = 2*PI-a1; end
   if s2 < 0 then a2 = 2*PI-a2; end 
   
   if math.abs(a1)+math.abs(a2) > 2*PI then 
     a1=-(2*PI-a1); a2=-(2*PI-a2) 
   end  

  
   local m1,m2 
   if bool(nedges) then
       m1 = math.ceil(math.abs( a1*nedges/(math.abs(a1)+math.abs(a2)) ))       
       m2 = math.ceil(math.abs( a2*nedges/(math.abs(a1)+math.abs(a2)) ))
       if m1>m2 then m2=nedges-m1 else m1=nedges-m2 end      
   elseif (self._edgelength>0) then
       m1 = math.ceil(math.abs(a1*r/self._edgelength))  
       m2 = math.ceil(math.abs(a2*r/self._edgelength))  
   else
       m1 = math.ceil(math.abs(a1*18/PI))       
       m2 = math.ceil(math.abs(a2*18/PI))       
   end     
   m1 = math.max(1,m1)
   m2 = math.max(1,m2)
   
   local p,p_old,i,v,pti
   local rcosi,rsini,dt,t
   

  if debug then 
   p = self:lineto(center); self:lineto(pt2); self:moveto(p); self:lineto(pt3); self:moveto(p1);
  end--if

   -- 1st part: arc from pt1 to pt3


   rcosi = (pt1[1]-center[1])  -- r* cos( rotation angle )
   rsini = (pt1[2]-center[2])  -- r* sin( rotation angle )
    
   p = p1
   dt = a1/m1               
   t = 0
   for i=1,m1-1 do
      t = t+dt
      v = { math.cos(t), math.sin(t) } 
      pti = {rcosi*v[1] - rsini*v[2], 
             rsini*v[1] + rcosi*v[2]}  -- rotation  
             
      p_old = p
      p = self:newpoint({center[1] + pti[1], 
                         center[2] + pti[2]})
      self:newfacet{p_old,p}
   end
   p_old = p
   self:newfacet{p_old,p3}
   self._savedpoint  = p3


   -- 2nd part:  arc from pt3 to pt2

   rcosi = (pt3[1]-center[1])  -- r*cos(rotation angle)
   rsini = (pt3[2]-center[2])  -- r*sin(rotation angle)
    
   dt = a2/m2              
   p = p3
   t = 0
   for i=1,m2-1 do
      t = t+dt
      v = { math.cos(t), math.sin(t) }
      pti = {rcosi*v[1] - rsini*v[2], 
             rsini*v[1] + rcosi*v[2]}  -- rotation  
      p_old = p      
      p = self:newpoint({center[1] + pti[1], center[2] + pti[2]})
      self:newfacet{p_old,p}
   end
   p_old = p
   
   self:newfacet{p_old,p2}
   self._savedpoint  = p2
   return p2
  end -- function arc3to        



  --- \brief  Generate an arc (of a circle sector) around a given center;
  ---         start from the current pen position. 
  ---         This function works only in 2D.
  ---         
  --- \param  center Center of the circle. 
  ---         The radius is determined by the distance of the 
  ---         center from the current pen position.
  --- \param  angle Angle of the arc in degrees.  
  ---         If the angle is positive, then the arc is drawn counter-clockwise
  ---         (i.e. in mathematical positive sense). 
  ---         If the angle is negative, then the arc is drawn clockwise.
  ---         The range of the angle is not limited.
  --- \param  nedges Number edges approximating the curve. 
  ---         If this parameter is omitted, then its value will be determined from
  ---         a proposed edgelength, cf. method edgelength(). 
  ---
  --- \return Index of the final point.
  ---         The position of the final point corresponds
  ---         to the specified angle.
  ---
  --+ int arcaround(center,angle,nedges);
  ---
  --- Example:
  --- -- yinyang
  --- g:edgelength(0.1) 
  --- g:moveto{0,0}
  --- g:facetregion(1);  g:arcaround({1,0}, 180) 
  --- g:facetregion(2);  g:arcaround({3,0},-180)  
  --- g:facetregion(3);  g:arcaround({2,0}, 180) -- !! take a big number of edges here
  --- g:facetregion(3);  g:arcaround({2,0}, 180) -- !! take a big number of edges here 
  --- g:regionnumber(1); g:regionpoint{1,0}; 
  --- g:regionnumber(2); g:regionpoint{3,0}; 
  ---
  --- -- pseudo yinyang
  --- g:moveto{0,0}
  --- g:facetregion(1);  g:arcaround({0,1}, 90,15)
  --- g:facetregion(2);  g:arcaround({2,1},-90,15)
  --- g:facetregion(3);  g:arcaround({1,1},180,35) 
  --- g:facetregion(3);  g:arcaround({1,1},180,35)
  --- g:regionnumber(1); g:regionpoint{0,1}
  --- g:regionnumber(2); g:regionpoint{2,1}
  ---  
  function g:arcaround(center,angle,nedges)
   
   --
   --              center o
   --                    / \                                 
   --                   /   \                    
   --        radius  r /angle\                  
   --                 /       \r   r = radius = | center -  pt1 |      
   --                /         \                 
   --  source point o           *  final point pt2  
   --           pt1  \         /   
   --                 '-.___.-'                                              
   --    
   -- index of source point 
   local p1 = table.getn(self.points)  
   if self._savedpoint and self._savedpoint>0 and self._savedpoint<p1 then 
     p1 = self._savedpoint 
   end
   if p1<1 then -- there is no source point
     error("arcaround: ERROR. No source point found.")
     return 0
   end
   
   local psource = p1               -- save the source point index
   local closed = false             -- don't close the circle
   
   -- special cases: 
   
      -- If abs(angle) is not less than 360 degrees, 
      -- we generate a closed circle.
      -- But in all cases we move the pen to the the position
      -- which is given by the full specified angle.
   
   if math.abs(angle)>=360 then   
      closed = true                    -- generate a closed circle
      if angle >360 then
        a = angle - 360*math.floor(angle/360)  -- additional angle 
        p1 = self:arcaround(center,a,nedges)   -- recursive call
        angle =  360 - a                       -- remaining angle  
      elseif angle <-360 then
        a = angle + 360*math.floor(-angle/360) -- additional angle 
        p1 = self:arcaround(center,a,nedges)   -- recursive call
        angle = -360 - a                       -- remaining angle  
      end
   end

   -- generate an arc with the given angle; 
   -- starting point is p1
   
   local PI = 3.14159265358979323846264
   local a = PI*angle/180                      -- angle in rad  

   local pt1   = self.points[p1]               -- coordinates of p1
   local rcosi = pt1[1]-center[1]              -- r*cos (rotation angle)
   local rsini = pt1[2]-center[2]              -- r*sin (rotation angle)
   local r     = math.sqrt(rcosi^2+rsini^2)    -- radius
   
   local m
   if bool(nedges) then
       m = math.ceil(nedges)
   elseif (self._edgelength>0) then
       m = math.ceil(math.abs(a*r/self._edgelength))  
   else
       m = math.ceil(math.abs(a*18/PI))       
   end
   m = math.max(1,m)
   
   local dt= a/m
   local t = 0
   local p = p1
   local p_old
   local v

   for i=1,m do
      p_old = p  -- save the previous point index 
      t = t+dt   
            
      if i==m and closed then
        self:newfacet{p_old,psource}  -- close the circle
        p = p1                        -- move the pen to p1 
      else
        -- new point on the curve 
        v   = { math.cos(t), math.sin(t) } 
        p   = self:newpoint{ center[1] + rcosi*v[1] - rsini*v[2], 
                             center[2] + rsini*v[1] + rcosi*v[2] }
        self:newfacet{p_old,p}
      end
   end
   self._savedpoint = p
   return p 
  end--function arcaround 


  --- \brief Generate an ellipse or a circle.
  --- \param center Center (as table of coordinates) of the ellipse.
  --- \param halfaxes Table of halfaxes or a radius.
  --- \param inclin Inclination angle; default value: 0.
  --- \param nedges Number edges approximating the curve. The minimum is 3.
  ---        If the number of edges is omitted or 0, then the edge length is 
  ---        determined from self._edgelength.  
  --- \return Index of the last point, which is equal to the starting point.
  ---
  --+ int ellipse(center,halfaxes,inclin,nedges);  
  
  function g:ellipse(center,halfaxes,inclin,nedges)
    local axes
    if type(halfaxes)=="table" then 
      axes = halfaxes
    elseif type(halfaxes)=="number" then  
      axes = {halfaxes,halfaxes}
    else return 0
    end
       
    local i
    local PI = 3.14159265358979323846264
    local m
    if bool(nedges) and nedges>0 then
       m = math.ceil(nedges)
    elseif (self._edgelength>0) then
       m = math.ceil(math.abs(2*PI*math.max(axes[1],axes[2])/self._edgelength))  
    else
       m = 36       
    end
    m = math.max(3,m)
    
    local dt = 2*PI/m
    local inclina = inclin or 0
    local cosi = math.cos(PI*inclina/180)
    local sini = math.sin(PI*inclina/180)
    local cosdt = math.cos(dt)
    local sindt = math.sin(dt)
    local v={}
    local w={1,0}
    local p,p_old,p0
      v = {axes[1]*w[1],axes[2]*w[2]}
      pt= {center[1] + cosi*v[1]-sini*v[2], center[2] + sini*v[1]+cosi*v[2]}
      p0= self:newpoint(pt)
      p_old=p0
    for i=2,m do
      w = {cosdt*w[1]-sindt*w[2], sindt*w[1]+cosdt*w[2]}
      v = {axes[1]*w[1],axes[2]*w[2]}
      pt= {center[1] + cosi*v[1]-sini*v[2], center[2] + sini*v[1]+cosi*v[2]}
      p = self:newpoint(pt)
      self:newfacet{p_old,p}
      p_old=p
    end
    self:newfacet{p_old,p0}
    self._savedpoint = p0
    return p0
  end--function ellipse


  --- \brief Generate a circle.
  --- \param center Center of the circle.
  --- \param radius Radius.
  --- \param inclin Inclination angle; default: 0.
  --- \param nedges Number edges approximating the curve. The minimum is 3.
  ---        If the number of edges is omitted or 0, then the edge length is 
  ---        determined from self._edgelength.  
  --- \return Index of the last point, which is equal to the starting point.
  ---
  --+ int circle(center,halfaxes,inclin,nedges);  
  function g:circle(center,radius,inclin,nedges)
    return self:ellipse(center,radius,inclin,nedges)
  end

------------------------------------------------
  return g

end--function pgdescCreate

--+ };

------------------------------------------------------------------------------
---
--- \brief Create an instance of rgdesc ("rectangular grid")
---
--- \param eps  Indicates the  
---   spare distance used to set masks.
--+ rgdesc rgdescCreate( double eps);



--- \brief Lua class rgdesc
--- Create rectangular grid without referring to 
--- grid generators.
--+ class rgdesc {
--+  
                    local function inmask(points,idx,mask,eps,dim)
                       local i
                       for i=1,table.getn(idx) do
                          if dim>0 and (points[idx[i] ][1]<mask[1][1]-eps or  
                                        points[idx[i] ][1]>mask[2][1]+eps) then return false end 
                          if dim>1 and (points[idx[i] ][2]<mask[1][2]-eps or  
                                        points[idx[i] ][2]>mask[2][2]+eps) then return false end 
                          if dim>2 and (points[idx[i] ][3]<mask[1][3]-eps or  
                                        points[idx[i] ][3]>mask[2][3]+eps) then return false end 
                       end
                       return true
                    end -- function inmask


                    local function region(points,idx,masks,eps,dim,select_first_mask)
                       local reg=0
                       if not masks then return 0 end
                       for i=1,table.getn(masks) do
                          if inmask(points,idx,masks[i],eps,dim) then reg=masks[i].region 
                          end
                          if reg>0 and select_first_mask then break 
                          end
                       end
                       return reg
                    end

         --- \brief Check if a point set belongs to a special region 
         ---        and return the found region number.  
         --- \param points Table of coordinates corresponding 
         ---               to all points of the grid.
         --- \param idx    Table of indices of points which should be inside the region.
         --- \param mask   A single region descriptions (mask).
         --- \param eps    A small length for unsharp comparision.
         ---
  local function rgSpecialRestriction(points,idx,mask,eps,dim) 
        

         local a={0,0,0}
         local x,y,z = 0,0,0
         local i,j,k
         local nindices = table.getn(idx) 
                   
         if (mask.shape=="box") then
               return mask.region
            
         elseif (mask.shape=="ell") then
               a={0,0,0}
               if dim>0 then a[1] = (mask[2][1]-mask[1][1])/2 end
               if dim>1 then a[2] = (mask[2][2]-mask[1][2])/2 end
               if dim>2 then a[3] = (mask[2][3]-mask[1][3])/2 end
               for j=1,nindices do
                 x,y,z = 0,0,0
                 if a[1]>0 then x = (points[idx[j] ][1]-mask[1][1])/a[1]-1 end
                 if a[2]>0 then y = (points[idx[j] ][2]-mask[1][2])/a[2]-1 end
                 if a[3]>0 then z = (points[idx[j] ][3]-mask[1][3])/a[3]-1 end
                 if x^2+y^2+z^2 > 1 + eps then return 0 end
               end--for j 
               return mask.region
               
         elseif  (mask.shape=="cyl")or(mask.shape=="zcyl")
               or(mask.shape=="ycyl")or(mask.shape=="xcyl") 
         then
               a={0,0,0}
               if dim>0 then a[1] = (mask[2][1]-mask[1][1])/2 end
               if dim>1 then a[2] = (mask[2][2]-mask[1][2])/2 end
               if dim>2 then a[3] = (mask[2][3]-mask[1][3])/2 end
               
               if (mask.shape=="xcyl") then a[1] = 0 end
               if (mask.shape=="ycyl") then a[2] = 0 end
               if (mask.shape=="zcyl") then a[3] = 0 end
               if (mask.shape=="cyl")  then a[3] = 0 end
               
               for j=1,nindices do
                 x,y,z = 0,0,0
                 if a[1]>0 then x = (points[idx[j] ][1]-mask[1][1])/a[1]-1 end
                 if a[2]>0 then y = (points[idx[j] ][2]-mask[1][2])/a[2]-1 end
                 if a[3]>0 then z = (points[idx[j] ][3]-mask[1][3])/a[3]-1 end
                 if x^2+y^2+z^2 > 1 + eps then return 0 end
               end--for j 
               return mask.region
               
         else 
               error("rgdesc: Undefined mask shape " .. mask.shape)
               return 0
                 
         end--if 
  end--function

                    local function rgRegion(points,idx,masks,eps,dim,select_first_mask)
                       local reg=0
                       if not masks then return 0 end
                       local i,r
                       for i=1,table.getn(masks) do
                          if inmask(points,idx,masks[i],eps,dim) then 
                           
                            -- We have found a valid bounding box of a region;
                            -- now we look for special restrictions.
                            if masks[i].shape then
                              r= rgSpecialRestriction(points,idx,masks[i],eps,dim)      
                                -- Because of backward compatibility,
                                -- the variable reg must remain its value until
                                -- another valid region is found.       
                              if r>0 then 
                                reg=r 
                                -- We have found a valid region and could stop here. 
                                -- break 
                                if select_first_mask then break 
                                end  
                              end 
                            else -- the shape is a "box" 
                              reg= masks[i].region; 
                              if select_first_mask then break 
                              end  
                            end
                            
                          
                          end--if inmask
                       end--for i
                       return reg -- return 
                    end -- function region

--+  public:

function rgdescCreate(eps)
return {
   _cellmasks={},
   _facemasks={},
   _eps=eps,

   
--- \brief Add region number mask for cells
--- Multiple calls of  this function  are possible.
--- \param mask Mask to describe regionnumber
--- A mask is a Lua table like <pre> {{x0,y0},{x1,y1},reg=ireg}</pre>
--+ cellmask(mask);

   cellmask=function(self,mask)
               table.insert(self._cellmasks,mask)
            end,

--- \brief Add facet region number mask for facets
--- Multiple calls of  this function  are possible.
--- \param mask Mask to describe facet region
--- A mask is a Lua table like <pre> {{x0,y0},{x1,y1},reg=ireg}</pre>
--+ facemask(mask);

   facemask=function(self,mask)
               table.insert(self._facemasks,mask)
            end,

--- \brief Choice of the selection mode.
--- \param mode Selection mode: "first" or "last";
---        "first" means, that the search for a valid region 
---        always stops when the first valid one was found.
---        (This works similar to if-then-elseif-else.)
---        "last" means, that the search doesn't stop
---        before the last mask was checked.
--+ selectionmode(mode);

   selectionmode=function(self,mode)
     if mode=="first" then  self._select_first_mask=true end  
     if mode=="last"  then  self._select_first_mask=false end  
   end,
   
--- \brief Set x coordinate array
--- \param coord Array describing coordinates
--+ xcoord(coord);

   xcoord=function(self,coord)
             self._xcoord=coord
          end,

--- \brief Set y coordinate array
--- \param coord Array describing coordinates
--+ ycoord(coord);

   ycoord=function(self,coord)
             self._ycoord=coord
          end,

--- \brief Set z coordinate array
--- \param coord Array describing coordinates
--+ zcoord(coord);


   zcoord=function(self,coord)
             self._zcoord=coord
          end,

--- \brief Set coordinate according to mask
--- \param mask Mask describing data, like
--- <pre> {{x0,y0},{x1,y1}, nx=10, ny=20} </pre>
--+ rect(mask);

   rect=function(self,mask)
           
           local function coord(gridmask,dim,n)
              if not n  or not gridmask[1][dim] then return nil end
              n=math.ceil(n)
              local crd={}
              local h=(gridmask[2][dim]-gridmask[1][dim])/(n-1)
              local x=gridmask[1][dim]
              for i=1,n do 
                 crd[i]=x
                 x=x+h
              end
              return crd
           end

           self:xcoord(coord(mask,1,mask.nx))
           self:ycoord(coord(mask,2,mask.ny))
           self:zcoord(coord(mask,3,mask.nz))
         end, -- function coord


--- \brief Define intervals with flipped edge orientations.
---
--- \param x Switching points in x direction. 
--- \param y Switching points in y direction.
--- \param z Switching points in z direction. 
--+ switches(xtable, ytable, ztable);

   switches=function(self,x,y,z)
     self._xswitches=x; 
     self._yswitches=y; 
     self._zswitches=z;
   end,


--- \brief Generate grid (new version)
--+ generateGrid();

   generateGrid=function(self)
   
                    
        local function isflip(x,s,i1)
          if not bool(s) then return 0,1 end
          local n = table.getn(s)
          local ii
          for ii=i1,n do 
            if x < s[ii] then 
              return math.mod(ii-1,2),ii
            end
          end
          return math.mod(n,2),1      
        end
                    
    local cache=250
    if self.cache then cache=self.cache end
    local dim
    
    if self._xcoord then dim=1 end
    if self._ycoord then dim=2 end
    if self._zcoord then dim=3 end

    -----------------------------------------------------------
                    
    if dim==1 then
       local nx=table.getn(self._xcoord)
       local np=nx
       local nc=(nx-1)
       local nb=0 
       local points={}
       local cells={}
       local rcell={}
       local faces={}
       local rface={}
       local i,j,k,ip,ic,ib
                       
       for i=1,nx  do
            points[i]={self._xcoord[i]}
       end

       ic = 0
       ib = 0
                       
       for i=1,nx-1  do
                
          p0 = i 
          p1 = i+1

          reg = rgRegion(points, {p0,p1}, self._cellmasks, self._eps, dim, self._select_first_mask)
          if reg>0 then ic=ic+1; cells[ic] = {p0,p1}; rcell[ic] = reg end 

          reg = rgRegion(points, {p0}, self._facemasks, self._eps, dim, self._select_first_mask)
          if reg>0 then ib=ib+1; faces[ib] = {p0}; rface[ib] = reg end 

        end--for i
        
        p1 = nx
        reg = rgRegion(points, {p1}, self._facemasks, self._eps, dim, self._select_first_mask)
        if reg>0 then ib=ib+1; faces[ib] = {p1}; rface[ib] = reg end 
                  
        return mgrid.gCreate(DArray2(points),
                            IArray2(cells), 
                            IArray1(rcell),
                            IArray2(faces),
                            IArray1(rface),
                            cache)
                              
    end--if dim==1
    -----------------------------------------------------------              
                    
    if dim==2 then
       local nx=table.getn(self._xcoord)
       local ny=table.getn(self._ycoord)
       local np=nx*ny
       local nc=2*(nx-1)*(ny-1)
       local nb=0 
       local points={}
       local cells={}
       local rcell={}
       local faces={}
       local rface={}
       local i,j,k,ip,ic,ib
       local xflip,yflip
       local izswitch,iyswitch
                       
       ip=0
       for j=1,ny do
         for i=1,nx  do
            ip = ip+1
            points[ip]={self._xcoord[i], self._ycoord[j]}
         end
       end
       
       ic = 0
       ib = 0
                       
       iyswitch = 1
       for j=1,ny-1 do
          yflip,iyswitch = isflip(self._ycoord[j],self._yswitches,iyswitch)
          ixswitch = 1
          for i=1,nx-1  do
            xflip,ixswitch = isflip(self._xcoord[i],self._xswitches,ixswitch)
                                
            ip = 1+(i-1)+(j-1)*nx
            --
            --  The 4 vertices of the box are:
            --  ^ dy
            --  |    01 --- 11 
            --  |     |     |  
            --  |     |     |   
            --  |    00 --- 10  
            --  +--------------> dx
            p00 = ip 
            p10 = ip+1
            p01 = ip  +nx
            p11 = ip+1+nx
                        
            if xflip==1 then 
              p00,p10 = p10,p00
              p01,p11 = p11,p01
            end
            if yflip==1 then 
             p00,p01 = p01,p00
             p10,p11 = p11,p10
            end
                                
            -- the box is splitted into 2 triangles:
            local c,ii
         
            c = { {p00,p10,p11}, {p11,p01,p00} }
         
            for ii=1,table.getn(c) do
              reg = rgRegion(points, c[ii], self._cellmasks, self._eps, dim, 
                self._select_first_mask)
              if reg>0 then 
                ic=ic+1; cells[ic] = c[ii]; rcell[ic] = reg 
              end 
            end

            -- there are 4 sides of the box
            local b
            b = { {p00,p10}, -- dy = 0
                  {p01,p11}, -- dy = 1
                  {p00,p01}, -- dx = 0
                  {p10,p11}} -- dx = 1

                local f={}
                if yflip==0 then  table.insert(f,b[ 1]);
                else              table.insert(f,b[ 2]);
                end
                if xflip==0 then  table.insert(f,b[ 3]);
                else              table.insert(f,b[ 4]);
                end

                if j==ny-1 then  
                  if yflip==1 then table.insert(f,b[ 1]);
                  else             table.insert(f,b[ 2]);
                  end
                end 
                if i==nx-1 then  
                  if xflip==1 then table.insert(f,b[ 3]);
                  else             table.insert(f,b[ 4]);
                  end
                end 
                     


            for ii=1,table.getn(f) do
              reg = rgRegion(points, f[ii], self._facemasks, self._eps, dim, 
                self._select_first_mask)
              if reg>0 then 
                ib=ib+1; faces[ib] = f[ii]; rface[ib] = reg 
              end 
            end
         
          end--for i
        end-- for j
                  
        return mgrid.gCreate(DArray2(points),
                            IArray2(cells), 
                            IArray1(rcell),
                            IArray2(faces),
                            IArray1(rface),
                            cache)
                              
    end--if dim==2
    -----------------------------------------------------------              
    if dim==3 then
       local nx=table.getn(self._xcoord)
       local ny=table.getn(self._ycoord)
       local nz=table.getn(self._zcoord)
       local nxy=nx*ny
       local np=nxy*nz
       local nc=6*(nx-1)*(ny-1)*(nz-1)
       local nb=0 
--       local points=mbase.da2Create(np,3)
--       local cells =mbase.ia2Create(nc,4)
--       local rcell =mbase.ia1Create(nc)
       local points={}
       local cells={}
       local rcell={}
       local faces={}
       local rface={}
       local b={}
       local c={}
       local f={}
       local i,j,k,ip,ic,ib,ii
       local xflip,yflip,zflip
       local izswitch,iyswitch,ixswitch
                       
       ip=0
       for k=1,nz do
          for j=1,ny do
             for i=1,nx  do
                ip = ip+1
--              mbase.da2Set(points,ip,1,self._xcoord[i])
--              mbase.da2Set(points,ip,2,self._ycoord[j])
--              mbase.da2Set(points,ip,3,self._zcoord[k])
                points[ip]={self._xcoord[i], self._ycoord[j], self._zcoord[k]}
             end
          end
       end
 
       ic = 0
       ib = 0
       izswitch = 1                     
       for k=1,nz-1 do
          zflip,izswitch = isflip(self._zcoord[k],self._zswitches,izswitch)
          iyswitch = 1
          for j=1,ny-1 do
             yflip,iyswitch = isflip(self._ycoord[j],self._yswitches,iyswitch)
             ixswitch = 1
             for i=1,nx-1  do
                xflip,ixswitch = isflip(self._xcoord[i],self._xswitches,ixswitch)
                
                ip = 1+(i-1)+(j-1)*nx+(k-1)*nxy
                --
                --  The 8 vertices of the box are:
                --
                --    001 ---- 011 
                --     |\       .\
                --     | \      . \
                --     | 101 ---- 111
                --     |  |     .  |
                --    000.|....010 | 
                --      \ |      . |   
                --       \|       .|
                --       100 ---- 110
                --

                p000 = ip 
                p100 = ip+1
                p010 = ip  +nx
                p110 = ip+1+nx
                p001 = ip     +nxy
                p101 = ip+1   +nxy
                p011 = ip  +nx+nxy
                p111 = ip+1+nx+nxy
                
                -- The following flipping of points is a trick to reduce 
                -- the number of code lines.  
                -- TODO: find a better trick!
        
                if xflip==1 then 
                  p000,p100 = p100,p000
                  p010,p110 = p110,p010
                  p001,p101 = p101,p001
                  p011,p111 = p111,p011
                end
                if yflip==1 then 
                  p000,p010 = p010,p000
                  p100,p110 = p110,p100
                  p001,p011 = p011,p001
                  p101,p111 = p111,p101
                end
                if zflip==1 then 
                  p000,p001 = p001,p000
                  p100,p101 = p101,p100
                  p010,p011 = p011,p010
                  p110,p111 = p111,p110
                end

                -- the box is splitted into 6 tetrahedra:
                                
                c = { {p000,p100,p110,p111},
                      {p000,p100,p101,p111},
                      {p000,p010,p011,p111},
                      {p000,p010,p110,p111},
                      {p000,p001,p101,p111},
                      {p000,p001,p011,p111}}
         
                for ii=1,table.getn(c) do
                  reg = rgRegion(points, c[ii], self._cellmasks, self._eps, dim, 
                    self._select_first_mask)
                  if reg>0 then 
                    ic=ic+1; cells[ic] = c[ii]; rcell[ic] = reg 
                  end 
                end
         
                -- there are 6 rectangular sides of the box
                -- each side is splitted into 2 triangular facets; 
                        
                b = { {p000,p010,p110}, {p000,p100,p110},  -- dz==0
                      {p001,p011,p111}, {p001,p101,p111},  -- dz==1
                      {p000,p001,p101}, {p000,p100,p101},  -- dy==0 
                      {p010,p011,p111}, {p010,p110,p111},  -- dy==1
                      {p000,p001,p011}, {p000,p010,p011},  -- dx==0
                      {p100,p101,p111}, {p100,p110,p111}}  -- dx==1

                -- usually, we insert only the triangles belonging to 3 sides of the box
                -- exception are made at the boundary of the domain.
                -- Furthermore, possible flippings of edge directions have to be
                -- taken into account.
                local f={}
                if zflip==0 then  table.insert(f,b[ 1]); table.insert(f,b[ 2]);
                else              table.insert(f,b[ 3]); table.insert(f,b[ 4]);
                end
                if yflip==0 then  table.insert(f,b[ 5]); table.insert(f,b[ 6]);
                else              table.insert(f,b[ 7]); table.insert(f,b[ 8]);
                end
                if xflip==0 then  table.insert(f,b[ 9]); table.insert(f,b[10]);
                else              table.insert(f,b[11]); table.insert(f,b[12]);
                end

                if k==nz-1 then  
                  if zflip==1 then table.insert(f,b[ 1]); table.insert(f,b[ 2]);
                  else             table.insert(f,b[ 3]); table.insert(f,b[ 4]);
                  end
                end 
                if j==ny-1 then  
                  if yflip==1 then table.insert(f,b[ 5]); table.insert(f,b[ 6]);
                  else             table.insert(f,b[ 7]); table.insert(f,b[ 8]);
                  end
                end 
                if i==nx-1 then  
                  if xflip==1 then table.insert(f,b[ 9]); table.insert(f,b[10]);
                  else             table.insert(f,b[11]); table.insert(f,b[12]);
                  end
                end 
                     
                for ii=1,table.getn(f) do
                  reg = rgRegion(points, f[ii], self._facemasks, self._eps, dim, 
                    self._select_first_mask)
                  if reg>0 then 
                    ib=ib+1; faces[ib] = f[ii]; rface[ib] = reg 
                  end 
                end
                 
             end--for i
          end-- for j
       end-- for k

       return mgrid.gCreate(
           DArray2(points),
           IArray2(cells), 
           IArray1(rcell),
           IArray2(faces),
           IArray1(rface),
           cache)
                       
    end --if dim==3



  end, --function generate                     
        
  --------------------------------------------------------------------------------------------
--- \brief Generate grid (older version)
--+ generate(newversion);
   generate=function(self,newversion)
   
      if bool(newversion) then 
        return self:generateGrid() 
      end 
      if bool(self._xswitches) or  
         bool(self._yswitches) or  
         bool(self._zswitches)   
      then       
        return self:generateGrid() 
      end 

                    local cache=250
                    if self.cache then cache=self.cache end
                    local dim
                    
                    if self._xcoord then dim=1 end
                    if self._ycoord then dim=2 end
                    if self._zcoord then dim=3 end


  --------------------------------------------------------------------------------------------

                    if dim==1 then
                       local nx=table.getn(self._xcoord)
                       local np=nx
                       local nc=nx-1
                       local nb=2
                       local points={}
                       local cells={}
                       local faces={}
                       local rcell={}
                       local rface={}
                       local i,j,k,ip,ib,ic
                       ip=1
                       ic=1
                       ib=1
                       
                       for i=1,nx  do
                          points[ip]={self._xcoord[i]}
                          if i<nx then
                             cells[ic]={ip,ip+1}
                             ic=ic+1
                          end
                          ip=ip+1
                       end
                       
                       ic=1
                       ip=1
                       for i=1,nx  do
                          if  i<nx then
                             local reg=region(points,{ip,ip+1},
                               self._cellmasks,self._eps,dim, self._select_first_mask)
                             rcell[ic]=reg
                             ic=ic+1
                          end
                          ip=ip+1
                       end
                       
                       
                       faces[ib]={1}
                       rface[ib]=region(points,faces[ib],
                         self._facemasks, self._eps,dim, self._select_first_mask)
                       ib=ib+1

                       faces[ib]={nx}
                       rface[ib]=region(points,faces[ib],
                         self._facemasks,self._eps,dim, self._select_first_mask)
                       ib=ib+1
                       return gCreate(DArray2(points),
                                      IArray2(cells), 
                                      IArray1(rcell),
                                      IArray2(faces),
                                      IArray1(rface),
                                      cache)
                    end -- if dim==1
                    
  --------------------------------------------------------------------------------------------
                    if dim==2 then
                       local nx=table.getn(self._xcoord)
                       local ny=table.getn(self._ycoord)
                       local np=nx*ny
                       local nc=2*(nx-1)*(ny-1)
                       local nb=2*(nx-1)+2*(ny-1)
                       local points={}
                       local cells={}
                       local faces={}
                       local rcell={}
                       local rface={}
                       local i,j,k,ip,ic,reg
                       
                       ip=1
                       ic=1
                       ib=1
                       
                       for j=1,ny do
                          for i=1,nx  do
                             points[ip]={self._xcoord[i],self._ycoord[j]}
                             if j<ny and i<nx then
                                cells[ic+0]={ip, ip   +1, ip+nx+1}
                                cells[ic+1]={ip, ip+nx+1, ip+nx  }
                                ic=ic+2
                             end
                             ip=ip+1
                          end
                       end
                       
                       ic=1
                       ip=1
                       for j=1,ny do
                          for i=1,nx  do
                             if j<ny and i<nx then
                                reg=region(points,{ip,ip+nx+1},
                                  self._cellmasks,self._eps,dim, self._select_first_mask)
                                rcell[ic+0]=reg
                                rcell[ic+1]=reg
                                ic=ic+2
                             end
                             ip=ip+1
                          end
                       end
                       
                       for i=1,nx-1 do
                          faces[ib]={i,i+1}
                          rface[ib]=region(points,faces[ib],
                            self._facemasks,self._eps,dim, self._select_first_mask)
                          ib=ib+1
                       end
               
                       for i=nx*(ny-1)+1, nx*ny-1 do
                          faces[ib]={i,i+1}
                          rface[ib]=region(points,faces[ib],
                            self._facemasks,self._eps,dim, self._select_first_mask)
                          ib=ib+1
                       end
                       
                       for j=1,nx*(ny-1),nx do
                          faces[ib]={j,j+nx}
                          rface[ib]=region(points,faces[ib],
                            self._facemasks,self._eps,dim, self._select_first_mask)
                          ib=ib+1
                       end
                       
                       for j=nx,nx*(ny-1),nx do
                          faces[ib]={j,j+nx}
                          rface[ib]=region(points,faces[ib],
                            self._facemasks,self._eps,dim, self._select_first_mask)
                          ib=ib+1
                       end
                       
                       return mgrid.gCreate(DArray2(points),
                              IArray2(cells), 
                              IArray1(rcell),
                              IArray2(faces),
                              IArray1(rface),
                              cache)
                    end -- if dim==2
  --------------------------------------------------------------------------------------------

                    if dim==3 then
                       local nx=table.getn(self._xcoord)
                       local ny=table.getn(self._ycoord)
                       local nz=table.getn(self._zcoord)
                       local nxy=nx*ny 
                       local np=nx*ny*nz
                       local nc=6*(nx-1)*(ny-1)*(nz-1)
                       local nb=4*((nx-1)*(ny-1)+(nz-1)*(ny-1)+(nz-1)*(nx-1))
                       local points=mbase.da2Create(np,3)
                       local cells =mbase.ia2Create(nc,4)
                       local faces =mbase.ia2Create(nb,3)
                       local rcell =mbase.ia1Create(nc)
                       local rface =mbase.ia1Create(nb)
                       local i,j,k,ip,ic,ib,reg
                       local node000,node100,node010,node110,node001,node101,node011,node111
                       local node00,node10,node01,node11
                       
                       local function ixset(a,idx,tab) 
                         local n=table.getn(tab)
                         local ii
                         for ii=1,n do mbase.ia2Set(a,idx,ii,tab[ii]) 
                         end
                       end
                       
                       local function  node(i,j,k) 
                         return 1+(i-1)+(j-1)*nx+(k-1)*nxy 
                       end
                       
                       ip=1
                       for k=1,nz do
                          for j=1,ny do
                             for i=1,nx  do
                                mbase.da2Set(points,ip,1,self._xcoord[i])
                                mbase.da2Set(points,ip,2,self._ycoord[j])
                                mbase.da2Set(points,ip,3,self._zcoord[k])
                                ip=ip+1
                             end
                          end
                       end
                       
                       ic=1
                       for k=1,nz-1 do
                          for j=1,ny-1 do
                             for i=1,nx-1  do
                             
                                node000= node(i+0,j+0,k+0) 
                                node100= node(i+1,j+0,k+0)
                                node010= node(i+0,j+1,k+0) 
                                node110= node(i+1,j+1,k+0)
                                node001= node(i+0,j+0,k+1) 
                                node101= node(i+1,j+0,k+1)
                                node011= node(i+0,j+1,k+1) 
                                node111= node(i+1,j+1,k+1)
                                
                                ixset(cells,ic+0,{node000,node100,node110,node111})                             
                                ixset(cells,ic+1,{node000,node100,node111,node101})                             
                                ixset(cells,ic+2,{node001,node101,node111,node000})
                                ixset(cells,ic+3,{node001,node011,node111,node000})
                                ixset(cells,ic+4,{node000,node010,node111,node011})
                                ixset(cells,ic+5,{node000,node110,node010,node111})
                                
                                reg=region(points, {node000,node111}, 
                                  self._cellmasks,self._eps,dim, self._select_first_mask)
                                
                                mbase.ia1Set(rcell,ic+0,reg)
                                mbase.ia1Set(rcell,ic+1,reg)
                                mbase.ia1Set(rcell,ic+2,reg)
                                mbase.ia1Set(rcell,ic+3,reg)
                                mbase.ia1Set(rcell,ic+4,reg)
                                mbase.ia1Set(rcell,ic+5,reg)

                                ic=ic+6
                             end--i
                          end--j
                       end--k
                       
                       -- boundary facets
                       ib=1
                       -- parallel to x-y plane
                       for k=1,nz,nz-1 do 
                       for j=1,ny-1 do 
                       for i=1,nx-1 do 
                         node00= node(i+0,j+0,k)
                         node10= node(i+1,j+0,k)
                         node01= node(i+0,j+1,k)
                         node11= node(i+1,j+1,k)
                         ixset(faces,ib+0,{node00,node10,node11})                               
                         ixset(faces,ib+1,{node11,node01,node00})                               
                         reg=region(points, {node00,node11}, 
                           self._facemasks,self._eps,dim, self._select_first_mask)
                         mbase.ia1Set(rface,ib+0,reg)
                         mbase.ia1Set(rface,ib+1,reg)
                         ib=ib+2
                       end
                       end
                       end
                       
                       -- parallel to y-z plane
                       for i=1,nx,nx-1 do 
                       for k=1,nz-1 do 
                       for j=1,ny-1 do 
                         node00= node(i,j+0,k+0)
                         node10= node(i,j+1,k+0)
                         node01= node(i,j+0,k+1)
                         node11= node(i,j+1,k+1)
                         ixset(faces,ib+0,{node00,node10,node11})                               
                         ixset(faces,ib+1,{node11,node01,node00})                               
                         reg=region(points, {node00,node11}, 
                           self._facemasks,self._eps,dim, self._select_first_mask)
                         mbase.ia1Set(rface,ib+0,reg)
                         mbase.ia1Set(rface,ib+1,reg)
                         ib=ib+2
                       end
                       end
                       end
                       
                       -- parallel to x-z plane
                       for j=1,ny,ny-1 do 
                       for k=1,nz-1 do 
                       for i=1,nx-1 do 
                         node00= node(i+0,j,k+0)
                         node10= node(i+1,j,k+0)
                         node01= node(i+0,j,k+1)
                         node11= node(i+1,j,k+1)
                         ixset(faces,ib+0,{node00,node10,node11})                               
                         ixset(faces,ib+1,{node11,node01,node00})                               
                         reg=region(points, {node00,node11}, 
                           self._facemasks,self._eps,dim, self._select_first_mask)
                         mbase.ia1Set(rface,ib+0,reg)
                         mbase.ia1Set(rface,ib+1,reg)
                         ib=ib+2
                       end
                       end
                       end
                       
                       return mgrid.gCreate(
                              DArray2(points),
                              IArray2(cells), 
                              IArray1(rcell),
                              IArray2(faces),
                              IArray1(rface),
                              cache)
                                   
                    end -- if dim==3
                 end, -- function generate
                 
     }   
  end -- function rgdescCreate

--+ };  

------------------------------------------------------------------------------------------------------------------
---
--- \brief 1D graded grid
---
local function g1d_adapt_b(bstart, n, r)

  --// Using Newton's method, 
  --// we determine b>0 such that f(b) = r, 
  --// where
  --//   f(b) := 1 + b + b^2 + ... + b^(n-2)
  --// For b>0, the function f is positive and monotone increasing.
  --// It holds  f(1) = n-1
  --// and       f(b) = (b^(n-1)-1)/(b-1), if b!=1.
  --// 
  --//  Newton's method:   bnew = b - (f(b)-r)/fprime(b)
  
  local j = 1 
  local maxiter = 1000
  local e = 0
  local eps = 1e-15 
  local b = bstart
  local bpow 
  
  if (r == n-1) then return 1 end

  for j=1,maxiter do
     if (1+eps < b or b < 1-eps) then  
      bpow = math.pow(b,n-2)
      -- f = (bpow*b -1)/(b-1),  
      -- fprime = (bpow*(n-1) - f)/(b-1),
      -- e = (f-r)/fprime;
      e = ( (bpow*b -1) - r*(b-1) )*(b-1)/( bpow*(n-1)*(b-1) - (bpow*b -1) );
    else            
      -- f = n-1, 
      -- fprime = (n-1)*(n-2)/2.0,
      -- e = (f-r)/fprime;
      e = 2*(n-1 - r)/((n-1)*(n-2));  
    end--if     
    b = b - e;
    if (-eps<e and e<eps) then break end
  end--do
  return b;
end--function


---\brief Generation of a graded 1D-grid 
---       with exponential growth of interval length. 
---             
---\param    xfirst,  first coordinate
---\param    xlast,   last coordinate
---\param    hfirst,  proposed length of the first discretisation interval or 0
---\param    hlast,   proposed length of the last  discretisation interval or 0
---\param    n        proposed number of discretisation points or 0;
---                   at least, n or one of the parameters hfirst, hlast has to be 
---                   specified.  
---\param    x        Optional table of coordinates; the
---                   generated coordinates will be appended to this table.
---\return  computed coordinates (if x is nil)
--- 
--+ g1dExpGraded(xfirst,xlast,  hfirst,hlast,n,x); 

function g1dExpGraded ( xfirst, xlast, hfirst, hlast, n, x )
    
    --   Ansatz:   x[i] = x[i-1] + hfirst * b^(i-2),   i=2,3,...,n
    --   Let be L = abs(xlast - xfirst).
    --   We obtain
    --   (1)     hlast   = x[n] - x[n-1] 
    --                   = hfirst * b^(n-2)
    --   (2)     L       = x[n] - x[1]   
    --                   = hfirst * (b^(n-1) - 1)/(b-1)
    --   (3)     L       = hlast  * (b^(n-1) - 1)/((b-1)*b^(n-2))
    --   (4)     L       = (hlast*b - hfirst )/(b-1)
    --  This means,
    --       b = (L - hfirst)/(L - hlast)
    --  
  local xlength = xlast - xfirst
  local L = math.abs(xlength)
  hfirst= math.abs(hfirst)
  hlast = math.abs(hlast)
  local h = hfirst
  local b = 1
  local i = 1 
  local xx = x or {}
  local eps = 1e-14
  
  --  Check what is given:
  if L <= hfirst then hfirst = 0 end
  if L <= hlast  then hlast  = 0 end
  if (n < 2) then
    if (hfirst > 0 and hlast > 0) then
      b= (L - hfirst)/(L - hlast) 
      if (b ~= 1) then n = 2 + math.ceil( math.log(hlast/hfirst)/math.log(b))
      else             n = 1 + math.ceil(L/(hfirst))
      end
    else
      b = 1;
      if (hfirst > 0)    then n = 1 + math.ceil(L/(hfirst))
      elseif (hlast > 0) then n = 1 + math.ceil(L/(hlast))
      else n = 2
      end
    end--if
  end--if
  if (hfirst > 0) then  
    b = g1d_adapt_b(b,n,L/hfirst) 
  elseif (hlast > 0) then
    b = 1/g1d_adapt_b(b,n,L/hlast) 
    hfirst= L - (L - hlast)*b
  else 
    b = 1 
    hfirst = L/(n-1)
  end--if
  h = hfirst
  if xlength<0 then h = -h end
  local nn = table.getn(xx)
  if nn>0 and math.abs(xx[nn]-xfirst)<eps*L then nn = nn-1 
  end
  xx[nn+1] = xfirst
  xx[nn+n] = xlast
  -- The first and last coordinates are explicitely set.
  for i=2,n-1 do xx[nn+i] = xx[nn+i-1] + h; h = h*b 
  end--do     
  return xx
end--function



---\brief Generation of a graded 1D-grid 
---       with linear  growth of interval length. 
---\param    xfirst,  first coordinate
---\param    xlast,   last coordinate
---\param    hfirst,  proposed length of the first discretisation interval or 0
---\param    hlast,   proposed length of the last  discretisation interval or 0
---\param    n,       proposed number of discretisation points or 0;
---                   at least, n or one of the parameters hfirst, hlast has to be 
---                   specified.  
---\param    x        Optional table of coordinates. 
---                   Geneated coordinates will be appended to this table.
---\return  computed coordinates (if x is nil)
---
--+ g1dLinGraded(xfirst,xlast,  hfirst,hlast,n,x); 

--- TODO: random test

function g1dLinGraded(xfirst,xlast,hfirst,hlast,n,x)
  --
  -- Ansatz:
  -- x[i+1] = x[i] + (i-1)*a + hfirst
  -- Therefore
  -- (1) hlast = (n-2)*a + hfirst
  -- (2) L     = (n-2)*(n-1)*a/2 + hfirst*(n-1)
  --     L     = (n-1)*(hlast + hfirst)/2 
  
  local xlength = xlast - xfirst
  local L=math.abs(xlength)
  hfirst= math.abs(hfirst)
  hlast = math.abs(hlast)
  local i
  local h
  local a
  local xx= x or {}
  local eps = 1e-14

  --  Check what is given:
  if L <= hfirst then hfirst = 0 end
  if L <= hlast  then hlast  = 0 end

  if n<2 then
    if hfirst>0 and hlast>0 then n=1+math.ceil(2*L/(hfirst+hlast)) 
    elseif hfirst>0 then n=1+math.ceil(L/(hfirst))
    elseif hlast >0 then n=1+math.ceil(L/(hlast))
    else n=2
    end
  end
  
    if hfirst>0 then hlast= 2*L/(n-1)-hfirst
    elseif hlast >0 then hfirst=2*L/(n-1)-hlast
    else hlast=L/(n-1); hfirst=hlast;
    end    
  
  if n>2 then a = (hlast-hfirst)/(n-2) else a=0 
  end
  h = hfirst
  if xlength<0 then h = -h; a=-a 
  end
  nn=table.getn(xx)
  if nn>0 and math.abs(xx[nn]-xfirst)<eps*L then nn=nn-1 
  end
  xx[nn+1]=xfirst
  xx[nn+n]=xlast
  for i=2,n-1 do xx[nn+i] = xx[nn+i-1] + h; h = h+a 
  end--do     
  return xx
end

---\brief Extend a table of 1D grid points.
---\param x Existing table.
---\param xx New table.
---\param ifirst Optional parameter: First index in table xx.
---       The default value is 2 if the last member of x 
---       and the first one of xx are identical; 
---       otherwise it is 1.
---
--+ g1dExtend(x,xx,ifirst); 

function g1dExtend(x,xx,ifirst)
  local n = table.getn(x)
  local nn= table.getn(xx)
  local eps = 1e-14
  local i1= ifirst or 1
  
  if (not bool(ifirst)) 
     and n*nn>0 
     and math.abs(x[n]-xx[1])<eps*math.abs(xx[nn]-xx[1]) then 
     i1 = 2 
  end
  for i=i1,nn do n=n+1; x[n]=xx[i] end
end

---\brief Generation of a graded 1D-grid 
---       with exponential growth of interval length. 
---             
---\param    xfirst,  first coordinate
---\param    xlength, length of the interval (may be negative)
---\param    hfirst,  proposed length of the first discretisation interval or 0
---\param    hlast,   proposed length of the last  discretisation interval or 0
---\param    n        proposed number of discretisation points or 0;
---                   at least, n or one of the parameters hfirst, hlast has to be 
---                   specified.  
---\param    x        Optional table of coordinates. 
---                   Geneated coordinates will be appended to this table.
---\return  computed coordinates (if x is nil)
--- 
--+ g1dGraded(xfirst,xlength,hfirst,hlast,n,x); 

function g1dGraded ( xfirst, xlength, hfirst, hlast, n, x )
print([[
---       Attention:
---       The function g1dGraded will be replaced by a more general one (TO DO !)
---       and its parameter list will change (!!!). 
---       Please use g1dExpGraded, g1dLinGraded, etc. instead !    
]])
  return g1dExpGraded( xfirst, xfirst+xlength, hfirst, hlast, n, x )
end--function g1dGraded
     
------------------------------------------------------------------------------------------------------------------

--+ slice(g3table,g2triangleoutput,z,zcellregions,xyfacetregions,zfacetregions);
                    
---\brief Generates a 3D-grid  from a given 2D-grid description g2 
---       and an array z of z-coordinates.
---\param g3table An extensible table containing the grid description. 
---       <pre> { points={...}, cells={...}, cellregions={...}, 
---               facets={...}, facetregions={...}} </pre>.
---       One or more slices can be put together in g3.
---\param g2triangleoutput A table containing the 2D-grid.
---\param z Table of z-coordinates. 
---\param zcellregions Table of cell-region numbers for the generated slice. 
---\param xyfacetregions Table of 2 tables corresponding to facet-region numbers 
---       on the bottom and the top of the generated slice. 
---\param zfacetregions Table of facet-region numbers for the generated slice. 

function slice(g3table,g2triangleoutput,z, 
                    zcellregions,xyfacetregions,zfacetregions)

 --  
 -- The original cell regions and facet regions of the 2D-grid
 -- are extruded as cellregions and facet regions of each generated slice.
 -- However, one can specify new region numbers using 
 -- zcellregions and zfacetregions. 
 -- Similar, the original cell regions can also be found
 -- as facet regions on the bottom and the top of the slice.
 -- New region numbers can be specified using "yzfacetregions[1]"
 -- for the boundaries at the bottom of the slice and "yzfacetregions[2]" 
 -- for the boundaries at the top.
 --

  local g3 = g3table
  local g2 = g2triangleoutput

  local npoints = da2NRow(g2.points)
  local ncells  = ia2NRow(g2.cells)
  local nfacets = ia2NRow(g2.facets)

  local ncellregions = 0 
  local nfacetregions = 0
  local nz = table.getn(z)
  local i,iz,iz1,izf,izf1,j,n1,n2,n3,n,ishift,iregion,iregionshift

  local npointslast = table.getn(g3.points)
  local ncellslast  = table.getn(g3.cells)
  local nfacetslast = table.getn(g3.facets)
  local ncellregionslast  = 0
  local nfacetregionslast = 0
  
  print(" ")
  print("slice: NEW SLICE from ",z[1]," to ",z[nz]) 
  print("slice:  number of x-y points =",npoints) 
  print("slice:  number of x-y cells  =",ncells)
  print("slice:  number of x-y facets =",nfacets)

  for i=1,ncells do 
   if ncellregions < g2.cellregions[i] 
    then ncellregions = g2.cellregions[i] 
   end 
  end
  for i=1,nfacets do 
   if nfacetregions < g2.facetregions[i] 
    then nfacetregions = g2.facetregions[i] 
   end 
  end

  print("slice:  number of x-y cellregions  =",ncellregions)
  print("slice:  number of x-y facetregions =",nfacetregions)
  
  print("slice: PREVIOUS SLICE:")
  print("slice:  npointslast =",npointslast) 
  print("slice:  ncellslast  =",ncellslast)
  print("slice:  nfacetslast =",nfacetslast)
  
  for i=1,ncellslast do 
   if ncellregionslast < g3.cellregions[i] 
    then ncellregionslast = g3.cellregions[i] 
   end 
  end
  for i=1,nfacetslast do 
   if nfacetregionslast < g3.facetregions[i] 
    then nfacetregionslast = g3.facetregions[i] 
   end 
  end

  print("slice:  ncellregionslast =",ncellregionslast)
  print("slice:  nfacetregionslast=",nfacetregionslast)

  local have_zcellregions  = bool(zcellregions)  and type(zcellregions)=="table"
  local have_zfacetregions = bool(zfacetregions) and type(zfacetregions)=="table" 
  local have_xyfacetregions= bool(xyfacetregions) 
                              and type(xyfacetregions)=="table" 
                              and type(xyfacetregions[1])=="table" 
                              and type(xyfacetregions[2])=="table" 

 
  if(npointslast>0) then iz1 = 2 else iz1 = 1  end -- first z-level
    
  for iz=iz1,nz do     -- loop over the z-levels
   ishift = npointslast+(iz-iz1)*npoints -- shift for the point index
   for j=1,npoints do 
     g3.points[j+ishift] = {g2.points[j][1],g2.points[j][2],z[iz]}
     --// print("point[j]",j+ishift,g3.points[j+ishift])
   end
  end

  i= ncellslast -- current cell index
  k= nfacetslast -- current facet index 
  
  iregionshift = ncellregionslast
  for iz=1,nz-1 do    -- loop over the z-levels
   ishift = npointslast + (iz-iz1)*npoints
   for j=1,ncells do
     n1 = g2.cells[j][1]+ishift
     n2 = g2.cells[j][2]+ishift
     n3 = g2.cells[j][3]+ishift
     if(n1>n2) then n=n1;n1=n2;n2=n; end
     if(n2>n3) then n=n2;n2=n3;n3=n; end
     if(n1>n2) then n=n1;n1=n2;n2=n; end
     g3.cells[i+1] ={n1,n2,n3,n3+npoints}
     g3.cells[i+2] ={n1,n2,n2+npoints,n3+npoints}
     g3.cells[i+3] ={n1,n1+npoints,n2+npoints,n3+npoints}
     iregion = iregionshift +g2.cellregions[j] -- automatic number 
     --// INSTEAD, WE CAN CHOOSE INDIVIDUAL REGION NUMBERS ON EACH LEVEL
     if have_zcellregions then
        iregion = zcellregions[g2.cellregions[j] ] 
      --iregion = zcellregions[iz][g2.cellregions[j]]
     end--if  
     g3.cellregions[i+1] = iregion
     g3.cellregions[i+2] = iregion
     g3.cellregions[i+3] = iregion
     i=i+3
   end--for j
   -- iregionshift = iregionshift + ncellregions
  end--for iz

  -- facets parallel to the x-y-plane

  iregionshift = nfacetregionslast
  
  if(npointslast>0) then izf1 = 2 else izf1 = 1  end 
  for izf = izf1,2 do 
   iz = 1
   if izf==2 then iz=nz end--if
   ishift = npointslast + (iz-iz1)*npoints
   for j=1,ncells do
    iregion = iregionshift + g2.cellregions[j]  -- automatic
    --// INSTEAD, WE CAN CHOOSE INDIVIDUAL REGION NUMBERS ON EACH LEVEL
    if have_xyfacetregions then
        iregion = xyfacetregions[izf][g2.cellregions[j] ]
    end--if  
    if iregion > 0 then
     n1 = g2.cells[j][1]+ishift
     n2 = g2.cells[j][2]+ishift
     n3 = g2.cells[j][3]+ishift
     g3.facets[k+1] = {n1,n2,n3}
     g3.facetregions[k+1]=iregion
     k=k+1
    end--if
   end--for j
   iregionshift = iregionshift + ncellregions
  end--for izf

  -- facets vertical to the x-y plane

  for iz=1,nz-1 do
   ishift = npointslast + (iz-iz1)*npoints
   for j=1,nfacets do 
     iregion = iregionshift + g2.facetregions[j]
     -- INSTEAD, WE CAN CHOOSE INDIVIDUAL REGION NUMBERS ON EACH LEVEL
     if have_zfacetregions then
       --iregion = zfacetregions[iz][g2.facetregions[j]]
       iregion = zfacetregions[g2.facetregions[j] ]
     end--if  
     n1=g2.facets[j][1]+ishift
     n2=g2.facets[j][2]+ishift
     if(n1>n2) then n=n1;n1=n2;n2=n; end
     g3.facets[k+1] = {n1,n2,n2+npoints}
     g3.facets[k+2] = {n1,n1+npoints,n2+npoints}
     g3.facetregions[k+1] = iregion
     g3.facetregions[k+2] = iregion
     k=k+2
   end
   --iregionshift = iregionshift + nfacetregions
  end--for iz
  
end -- function slice

------------------------------------------------------------------------------------------------------------------
-- TODO:  make a better API 
--        Current alpha version:
--        g = xgdescCreate(baseGrid)
--        g:genSlice(zcoord1,zcellregions1,xyfacetregions1,zfacetregions1)
--        g:genSlice(zcoord2,zcellregions2,xyfacetregions2,zfacetregions2)
--        ...
--- 
--- \brief Features for generation of extruded grids.
---        
---
--- \brief Create an instance of xgdesc (extruded grid description)
--- \param basegrid A triangular grid. 
---        Currently only the triangle output will be accepted.
--+ xgdesc xgdescCreate(basegrid);
---
--- \brief Lua class xgdesc.
---
--- Create an extruded grid slice by slice.
---
--+ class xgdesc { public:

function xgdescCreate(g2d)
 local xg = { points={}, cells={}, cellregions={}, 
              facets={}, facetregions={}}
              
 -- TODO: transform the 2D grid into a usual grid description.
              
 xg._triangulation = g2d   -- triangle(g2d,"paAq")            

 ---\brief Generate or extend a slice.
 ---\param z Table of z-coordinates. 
 ---\param zcellregions Table of cell-region numbers for the generated slice. 
 ---\param bottomfacetregions Table of facet-region numbers  
 ---       on the bottom of the generated slice. 
 ---\param topfacetregions Table of facet-region numbers  
 ---       on the top of the generated slice. 
 ---\param zfacetregions Table of facet-region numbers for the generated slice. 
 --+ genSlice(z,zcells,bottomfacets,topfacets,zfacets);
 
 function xg:genSlice(z, zcellregions, bottomfacetregions, topfacetregions, zfacetregions )
    slice(self,self._triangulation, z, zcellregions, 
    {bottomfacetregions,topfacetregions}, zfacetregions )
 end          
 
 return xg
end

--+}

------------------------------------------------------------------------------------------------------------------



